/*
	hoifix.c

	johan@linkdata.se
*/

#include <stdio.h>
#include <stdlib.h>

#define CRLF "\x0D\x0A"
#define HOIFIX_BUFSIZE 256*1024
char hexdigit[] = "0123456789ABCDEF";


int main( int argc, char **argv )
{
	FILE *in, *out;
	int linelen, linecount;
	unsigned char infilename[256];
	unsigned char outfilename[256];
	unsigned char *linebuf;
	unsigned char *last_line;
	unsigned char *p;
	unsigned char *first_quote, *last_quote;
	int quotecount;
	char replaced[256];

	setbuf(stdout, NULL);

	/* check we got a filename */
	if( argc < 2 )
	{
		in = fopen("autosave.eug","rb");
		if( in == NULL )
		{
			printf(
				"Hello, and welcome to HOIFIX!\n"
				"\n"
				"Now, if you read the README file, you should be running\n"
				"this program from where your Hearts of Iron save games\n"
				"are stored.\n"
				"\n"
				"However, I can't find an autosave.eug file in the current\n"
				"directory, so chances are you haven't done this. You can\n"
				"still continue by entering a filename to scan and fix, but\n"
				"if it's not in the current directory and you fail to provide\n"
				"a complete path to it, it won't be found.\n"
				"\n"
				);
		}
		else
		{
			fclose(in);
		}

		printf(
			"Please enter name of the file to scan, or press ENTER\n"
			"if you've changed your mind and want to abort. Remember\n"
			"to include the \".eug\" at the end of the file name.\n"
			"\n"
			"Filename: "
			);

		fgets( infilename, sizeof(infilename), stdin );
		linelen = strlen( infilename );
		while( linelen && infilename[linelen-1]<' ' ) infilename[--linelen] = 0;
		if( !linelen ) return 0;
	}
	else
	{
		strncpy( infilename, argv[1], sizeof(infilename) );
		infilename[ sizeof(infilename)-1 ] = 0;
	}

	/* open input */
	in = fopen( infilename, "rb" );
	if( in == NULL )
	{
		printf( "Can't open the file '%s'\n", infilename );
		return 2;
	}

	/* construct new filename */
	strcpy( outfilename, "FIXED-" );
	strncpy( outfilename+6, infilename, sizeof(outfilename)-7 );
	outfilename[sizeof(outfilename)-1] = 0;

	/* open output */
	out = fopen( outfilename, "wb" );
	if( out == NULL )
	{
		fclose( in );
		printf( "Can't create output file '%s'\n", outfilename );
		return 3;
	}

	linecount = 0;
	if( (linebuf = malloc(HOIFIX_BUFSIZE)) == NULL )
	{
		fclose( in );
		printf( "Out of memory\n" );
		return 9;
	}
	if( (last_line = malloc(HOIFIX_BUFSIZE)) == NULL )
	{
		free( linebuf );
		fclose( in );
		printf( "Out of memory\n" );
		return 9;
	}

	*last_line = 0;

	while( !feof(in) )
	{
		fgets( linebuf, HOIFIX_BUFSIZE, in );
		linelen = strlen(linebuf);
		linecount ++;
		if( linelen >= HOIFIX_BUFSIZE-2 )
		{
			printf( "Line %d: input buffer overflow; sorry, can't help you.\n", linecount );
			free( linebuf );
			free( last_line );
			fclose( in );
			fclose( out );
			remove( outfilename );
			return 4;
		}

		/* make sure that the line does not contain any */
		/* control characters except TABs or the CRLF at the end */
		quotecount = 0;
		p = linebuf;
		while( p < linebuf+linelen-2 )
		{
			if( *p < ' ' && *p != '\t' )
			{
				if( quotecount<63 )
				{
					replaced[quotecount*4+0] = hexdigit[(*p&0xF0)>>4];
					replaced[quotecount*4+1] = hexdigit[(*p&0x0F)];
					replaced[quotecount*4+2] = 'h';
					replaced[quotecount*4+3] = ' ';
				}
				quotecount ++;
				*p = ' ';
			}
			p ++;
		}
		if( quotecount )
		{
			replaced[quotecount*4-1] = 0;
			printf("Line %d: %d non-printable characters replaced with spaces (%s)\n", linecount, quotecount, replaced );
		}

		/*  check that all strings are terminated on the line */
		quotecount = 0;
		p = linebuf;
		while( p )
		{
			p = strchr( p, '\"' );
			if( p ) { quotecount ++; p ++; }
		}
		if( quotecount % 2 )
		{
			/* an uneven number of quotes. */

			if( *last_line )
			{
				/* if the last line also had an uneven number, we concatenate them. */
				printf("Line %d: merging this line with the previous line\n", linecount );
				fputs( last_line, out );
				*last_line = 0;
			}
			else
			{
				/* put this line on hold in last_line */
				memcpy( last_line, linebuf, HOIFIX_BUFSIZE );
				p = last_line + strlen(last_line);
				while( p >= last_line && *p < ' ' ) { *p = 0; p --; }
				continue;
			}
		}

		/* if last_line contains data, we're actually missing a quote */
		/* or there are one or more embedded quotes */
		if( *last_line )
		{
			/* this code assumes that a string won't contain '='. */
			/* not the best solution, but for the limited data */
			/* set found in HoI save games it works OK. */

			/* locate the extremes of the string(s), and replace all */
			/* doublequotes within with single quotes */

			p = last_line;
			quotecount = 0;
			while( *p )
			{
				/* outside string */
				last_quote = NULL;
				while( *p && *p != '\"' ) p++;
				if( !*p ) break;
				p ++;
				while( *p && *p != '=' )
				{
					if( *p == '\"' )
					{
						*p = '\'';
						quotecount ++;
						last_quote = p;
					}
					p ++;
				}
				if( last_quote )
				{
					quotecount --;
					*last_quote = '\"';
				}
				if( !*p )
				{
					if( last_quote == NULL )
					{
						/* just a single quote on that line. strange. */
						printf( "Line %d: added trailing double quote\n", linecount-1 );
						*(p++) = '\"';
						*p = 0;
					}
				}
			}

			if( quotecount )
				printf( "Line %d: replaced %d embedded double quotes with single quotes\n", linecount-1, quotecount );
			fputs( last_line, out );
			fputs( CRLF, out );
			*last_line = 0;
		}

		/* write the line */
		fputs( linebuf, out );
	}

	/* clean up and exit */
	free( linebuf );
	free( last_line );
	fclose( out );
	fclose( in );
	printf("Done writing '%s'. %d lines processed.\n", outfilename, linecount );

	/* ask to close if started w/o argument */
	if( argc < 2 )
	{
		printf( "Press ENTER to quit!\n" );
		fgets( infilename, sizeof(infilename), stdin );
	}

	return 0;
}
